'use strict';
const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const Table = require('cli-table3');
const { prompt, Select, Confirm } = require('enquirer');
const { ConnectionString } = require('connection-string');

const config = require('./config');

const cache = require('../lib/Cache');
const Auth = require('../lib/Auth');
const API = require('../lib/API');
const SSH = require('./sshconnect');

const hostPortRegex = /^(([a-zA-Z\-\.0-9]+):)?(\d+)$/;

const initCLI = async (argv) => {

	if(!config.has('appUserDir')) {
		const isWindows = (process.platform == 'win32');
		const homeDir = process.env[isWindows ? 'USERPROFILE' : 'HOME'];
		const appDir = path.join(homeDir,...(isWindows ? ['AppData', 'sshc'] : ['.sshc']));
		if(!fs.existsSync(appDir)) fs.mkdirSync(appDir);
		config.set('appUserDir', appDir);
	}

}

const connectWithoutAlias = async (args) => {
	const { _: [ connectionString ] } = args;

	const { user, hosts: [ serv ] } = new ConnectionString(connectionString);

	const dataConnect = {
		host: serv.name || 'localhost',
		port: serv.port || 22,
		username: user || 'root'
	}

	if (args.k) {
		try {
			const { data: prm } = await API.getKey(args.k);

			if(!prm || Object.keys(prm).length == 0) {
				console.log(chalk.red(`Nenhuma chave encontrada`));
				process.exit(0);
			}

			if(prm.pvkey) dataConnect.privateKey = prm.pvkey;
			if(prm.keypass) dataConnect.passphrase = prm.keypass;

		} catch (err) {
			console.error(chalk.red('Ocorreu um erro ao buscar chave...'));
			if(err.response && err.response.data && err.response.data.message) 
				console.error(chalk.red(err.response.data.message));
			else
				console.error(chalk.red(err.message || err.stack || err));
			process.exit(1);
		}
	} else if (args.key) {
		try {
			const choices = { 'Nenhuma': null };
			const keys = await API.listKeys();
			for (const k of keys.data) {
				choices[k.name] = k.id;
			}

			const selectedKey = await new Select({
				name: 'chave',
				message: 'Selecione uma chave:',
				choices: Object.keys(choices)
			}).run();

			if(selectedKey && !(selectedKey == 'Nenhuma' || typeof choices[selectedKey] == "undefined")) {
				const { data: prm } = await API.getKey(choices[selectedKey]);
				if(prm.pvkey) dataConnect.privateKey = prm.pvkey;
				if(prm.keypass) dataConnect.passphrase = prm.keypass;
			}

		} catch(err) {
			console.error(chalk.red('Ocorreu um erro ao buscar chaves...'));
			if(err.response && err.response.data && err.response.data.message) 
				console.error(chalk.red(err.response.data.message));
			else
				console.error(chalk.red(err.message || err.stack || err));
			process.exit(1);
		}
	} else {
		const formData = await prompt({
			type: 'password',
			name: 'password',
			message: 'Digite a senha do servidor (ou deixe em branco):'
		});
		if (formData.password) dataConnect.password = formData.password;

	}

	try {
		await SSH(dataConnect);
	} catch(err) {
		console.log(chalk.red(`Connect: Ocorreu um erro ao obter informações.`))
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		if(args.debug) console.log(err.stack || err);
		process.exit(1);
	}
}

const connect = async (args) => {
	const { _: [ alias ] } = args;

	if (/@/g.test(alias)) {
		return connectWithoutAlias(args);
	}

	try {
		const serverData = await API.skipCache(args.cache === false).getForConnect(alias).then(d => d.data);
		// console.log(serverData);
		const dataConnect = {
			host: serverData.host,
			port: serverData.port,
			username: serverData.username,
		}

		if(serverData.password) dataConnect.password = serverData.password;
		if(serverData.perm && serverData.perm.pvkey) dataConnect.privateKey = serverData.perm.pvkey;
		if(serverData.perm && serverData.perm.keypass) dataConnect.passphrase = serverData.perm.keypass;

		if(serverData.forwardOut && Object.keys(serverData.forwardOut).length > 0) {
			dataConnect.forwardOut = Object.values(serverData.forwardOut).map(f => {
				const a = f.split('|').map(ip => {
					const b = ip.split(':');
					return { host: b[0], port: +b[1] };
				});
				return { in: a[0], out: a[1] };
			});
		}

		if(serverData.forwardIn && Object.keys(serverData.forwardIn).length > 0) {
			dataConnect.forwardIn = Object.values(serverData.forwardIn).map(f => {
				const a = f.split(':');
				return { host: a[0], port: +a[1] };
			});
		}

		if(typeof args.hop == "undefined" && typeof args.proxy == "undefined" && typeof args.hopping == "undefined" && serverData.proxyServer && Object.keys(serverData.proxyServer).length > 0)
		{
			dataConnect.proxyServer = {
				host: serverData.proxyServer.host,
				port: serverData.proxyServer.port,
				username: serverData.proxyServer.username,
				alias: serverData.proxyServer.alias
			}
			if(serverData.proxyServer.password) dataConnect.proxyServer.password = serverData.proxyServer.password;
			if(serverData.proxyServer.perm && serverData.proxyServer.perm.pvkey) dataConnect.proxyServer.privateKey = serverData.proxyServer.perm.pvkey;
			if(serverData.proxyServer.perm && serverData.proxyServer.perm.keypass) dataConnect.proxyServer.passphrase = serverData.proxyServer.perm.keypass;
		}
		
		await SSH(dataConnect);

	} catch(err) {
		console.log(chalk.red(`Connect: Ocorreu um erro ao obter informações.`))
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		if(args.debug) console.log(err.stack || err);
		process.exit('SIGKILL');
	}
}

const clearCache = () => {
	return cache.clearAll();
}

const purge = () => {
	cache.clearAll();
	config.clear();
}

const debug = () => {
	console.log(config.get('tokenData'))
}

const validate = async () => {
	if(await Auth.validateSession())
		console.log(chalk.green.bold(`Sessão é válida!`));
	else
		console.log(chalk.red.bold(`Sessão invalida!`));
}

const refresh = async () => {
	try {
		const r = await Auth.refreshSession(true);
		if(r != true && r.data) {
			config.delete('tokenData');
			config.set('tokenData', r.data);
			console.log(chalk.green(`Sessão atualizada.`))
		} else {
			console.log(chalk.green(`Sessão ainda é valida.`));
		}
	} catch(err) {
		console.log(chalk.red(`Ocorreu um erro ao renovar sessão!`))
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
	}
}

const register = async () => {
	if(config.has('tokenData')) {
		console.log(chalk.bold(`Você já está logado`));
		console.log(chalk.bold(`Use sshc --logout`));
		process.exit(0);
	}

	if(!config.has('preRegisterEmail')) {

		const formData = await prompt([{
			type: 'input',
			name: 'username',
			message: 'Digite seu email:',
			validate: v => /^(([^<>()\[\]\\.,;:\s@"]+(\.[^<>()\[\]\\.,;:\s@"]+)*)|(".+"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$/.test(v),
			required: true
		},
		{
			type: 'password',
			name: 'password',
			message: 'Digite sua senha (min 8 caracteres):',
			required: true
		},
		{
			type: 'password',
			name: 'rpassword',
			message: 'Confirme sua senha:',
			required: true
		}]);

		if(formData.password != formData.rpassword) {
			console.log(chalk.red.bold(`Senhas não conferem, tente novamente.`));
			process.exit(1);
		}

		try {
			await Auth.register(formData.username, formData.password);
			console.log(chalk.green(`Cadastro efetuado!`));
			console.log(chalk.magenta(`Um codigo de confirmação foi enviado para o seu email!`));
			config.set('preRegisterEmail', formData.username);
		} catch(err) {
			console.error(chalk.red('Ocorreu um erro...'));
			if(err.response && err.response.data && err.response.data.message) 
				console.error(chalk.red(err.response.data.message));
			else
				console.log(chalk.red(err.message || 'unknown'));
			
			process.exit(1);
		}
	} else {
		console.log(chalk.magenta.bold(`Existe um cadastro em andamento...`));
		console.log(chalk.magenta.bold(`Um codigo de confirmação foi enviado para o seu email (${config.get('preRegisterEmail')})`));
	}

	const confirmCode = await prompt({
		type: 'numeral',
		name: 'code',
		message: 'Digite codigo de confirmação:',
		validate: v => !/[^0-9]+/g.test(v),
		required: true
	});
	
	try {
		await Auth.registerConfirm(config.get('preRegisterEmail'), confirmCode.code);
		console.log(chalk.green(`Cadastrado com sucesso! Digite 'sshc --login'`));
		config.delete('preRegisterEmail');
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

const login = async () => {

	if(config.has('tokenData')) {
		console.log(chalk.bold(`Você já está logado`));
		console.log(chalk.bold(`Use sshc --logout`));
		process.exit(1);
	}

	if(config.has('loginSessionMFA')) {
		loginMFA();
		return;
	}

	const formData = await prompt([{
		type: 'input',
		name: 'username',
		message: 'Digite seu email de login:',
		required: true
	},
	{
		type: 'password',
		name: 'password',
		message: 'Digite senha:',
		required: true
	}]);

	try {
		const r = await Auth.login(formData.username, formData.password);
		config.set('userEmail', formData.username);
		
		if(r.logged) {
			config.set('tokenData', r.data);
			console.log(chalk.green(`Login efetuado!`));
		} else if(r.data.nextStep) {
			switch(r.data.nextStep) {
				case 'MFA_AUTH':
					config.set('loginSessionMFA', r.data.loginSession);
					await loginMFA();
					break;
				default:
					console.log(chalk.red(`Unknown step (${r.data.nextStep})`));
					break;
			}
		}
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao fazer login...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

const relogin = async () => {

	if(!config.has('userEmail')) {
		console.error(chalk.red(`É necessário fazer o login novamente, tente 'sshc --login'`));
		process.exit(0);
	}

	const formData = await prompt({
		type: 'password',
		name: 'password',
		message: 'Digite senha:',
		required: true
	});

	try {
		const r = await Auth.login(config.get('userEmail'), formData.password);
		if(r.logged) {
			config.delete('tokenData');
			config.set('tokenData', r.data);
			console.log(chalk.green(`Login efetuado!`));
		} else if(r.data.nextStep) {
			switch(r.data.nextStep) {
				case 'MFA_AUTH':
					config.set('loginSessionMFA', r.data.loginSession);
					await loginMFA();
					break;
				default:
					console.log(chalk.red(`Unknown step (${r.data.nextStep})`));
					break;
			}
		}
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao fazer login...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		process.exit(1);
	}
}

const loginMFA = async () => {

	console.log(chalk.magenta.bold(`Autenticação multifator (MFA)\nFoi enviado um código para o número cadastrado.`));

	const formData = await prompt({
		type: 'numeral',
		name: 'loginCode',
		message: 'Digite o código recebido por SMS:',
		required: true
	});

	try {
		const r = await Auth.loginMFA(formData.loginCode);
		config.set('tokenData', r.data);
		console.log(chalk.green(`Login efetuado!`));
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao fazer login...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

const logout = async () => {
	if (!config.has('userEmail')) {
		return;
	}
	try {
		await Auth.logout();
		config.clear();
		await cache.clearAll();
		console.log(chalk.green(`Deslogado com sucesso!`));
	} catch (err) {
		console.error(chalk.red('Ocorreu um erro ao deslogar...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

const removePhoneNumber = async () => {
	try {
		const resp = await Auth.editPhoneNumber(null);
		if(resp.changed) console.log(chalk.green('Número removido com sucesso!'));
		else console.log(chalk.red('Não foi possivel remover.'));
	} catch(err) {
		console.log(chalk.red(`Ocorreu um erro...`))
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		process.exit('SIGKILL');
	}
}

const registerPhoneNumber = async (args) => {
	let phone = args.phone || null;

	if(!phone || phone === true) {
		const formData = await prompt({
			type: 'input',
			name: 'phone',
			message: 'Digite um número de celular Ex.: 551199999999',
			required: true
		}).then(d => {
			if(!d.phone.startsWith('+')) d.phone = '+'+d.phone;
			return d;
		});

		const confirmationPhone = await new Confirm({
			name: 'question',
			message: `Este número está correto? ${chalk.magenta(formData.phone)}`
		}).run();

		if(!confirmationPhone) {
			console.log(chalk.red(`Tente novamente.`));
			return false;
		}

		phone = formData.phone;
	}

	try {
		const resp = await Auth.editPhoneNumber(phone);
		if(resp.changed) console.log(chalk.green('Número cadastrado com sucesso!'));
		else console.log(chalk.red('Número não foi cadastrado.'), resp);
	} catch(err) {
		console.log(chalk.red(`Ocorreu um erro...`))
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		process.exit('SIGKILL');
	}
}

const statusMFA = async () => {

	try {
		const resp = await Auth.statusMFA();
		if(resp.statusMFA) {
			console.log(chalk.cyan.bold('MFA ativado para esta conta.'));
		} else {
			console.log(chalk.red.bold('MFA desativado para esta conta.'));
		}
	} catch(err) {
		console.log(chalk.red(`Ocorreu um erro ao obter informações.`))
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		process.exit('SIGKILL');
	}
}

const enableMFA = async () => {

	try {
		const resp = await Auth.changeStatusMFA(true);
		if(resp.changed)
			console.log(chalk.green.bold('MFA ativado com sucesso.'));
		else
			console.log(chalk.red.bold('Ocorreu um erro ao ativar.'));
	} catch(err) {
		if(err.response && err.response.data) {
			if(err.response.data.errCode == 'MissingPhoneNumber')
				console.log(chalk.cyan(`Para ativar MFA, você precisa cadastrar um número de celular a sua conta, tente '${chalk.bold('sshc --register-phone 551199999999')}'`));
			else
				console.log('Erro desconhecido');
		} else {
			console.log(chalk.red(`Ocorreu um erro...`))
			if(err.response && err.response.data && err.response.data.message) 
				console.error(chalk.red(err.response.data.message));
			else
				console.log(chalk.red(err.message || 'unknown'));
			process.exit('SIGKILL');
		}
	}
}

const disableMFA = async () => {

	try {
		const resp = await Auth.changeStatusMFA(false);
		if(resp.changed)
			if(resp.changedTo == false)
				console.log(chalk.green.bold('MFA desativado com sucesso.'));

	} catch(err) {
		console.log(chalk.red(`Ocorreu um erro...`))
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		process.exit('SIGKILL');
	}
}

const keyUnlink = async (args) => {
	const params = args._;

	if(!args.s && !args.g) {
		console.log(chalk.red.bold(`Uso incorreto, tente 'sshc key unlink [-s|-g] [server alias|group name]'`));
		process.exit(1);
	}

	if(args.s) {
		try {
			await API.unlinkKey(args.s, args.g);
			console.log(chalk.green.bold(`Chave desassociada com sucesso!`))
		} catch(err) {
			console.error(chalk.red('Ocorreu um erro ao desassociada chave...'));
			if(err.response && err.response.data && err.response.data.message) 
				console.error(chalk.red(err.response.data.message));
			else
				console.error(chalk.red(err.message || err.stack || err));
			
			process.exit(1);
		}
	}
}

const keyLink = async (args) => {
	const params = args._;

	if(!params[2]) {
		console.log(chalk.red.bold(`Uma chave deve ser especificada, tente 'sshc key link [name] [-s|-g] [server alias|group name]'`));
		process.exit(1);
	}

	if(!args.s && !args.g) {
		console.log(chalk.red.bold(`Uso incorreto, tente 'sshc key link [name] [-s|-g] [server alias|group name]'`));
		process.exit(1);
	}
	
	try {
		await API.linkKey(params[2], { server: args.s || null, group: args.g || null });
		console.log(chalk.green.bold(`Chave associada com sucesso!`))
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao associar chave...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

const keyImport = async (args) => {
	const params = args._;

	if(!params[2]) {
		console.log(chalk.red.bold(`Falta o caminho do arquivo! Tente 'sshc key import /path/to/private /path/to/public'`));
		process.exit(1);
	}

	if(!args.n) {
		console.log(chalk.red.bold(`Falta nome para a chave, tente 'sshc key import -n nomeChave ${params[2]}'`));
		process.exit(1);
	}

	if(/[^a-z0-9\.\-_]/gi.test(args.n)) {
		console.log(chalk.red.bold(`Nome inválido, é permitido apenas os caracteres [a-Z 0-9 - _ .]`));
		process.exit(1);
	}
	
	const formData = await prompt({
		type: 'password',
		name: 'password',
		message: 'Digite a senha da chave (ou deixe em branco):'
	});

	try {
		const name = args.n ? args.n : null;
		const privateFile = params[2] ? params[2] : null;
		const publicFile = params[3] ? params[3] : null;

		if(!privateFile || !fs.existsSync(privateFile)) {
			console.log(chalk.red(`O arquivo '${privateFile}' não existe.`));
			process.exit(1);
		}

		if(publicFile && !fs.existsSync(publicFile)) {
			console.log(chalk.red(`O arquivo '${publicFile}' não existe.`));
			process.exit(1);
		}
		
		const privateFileData = fs.readFileSync(privateFile).toString('utf8');
		const publicFileData = publicFile ? fs.readFileSync(publicFile).toString('utf8') : null;

		const r = await API.newKey(name, privateFileData, formData.password, publicFileData);

		console.log(chalk.green.bold(`Chave salva com sucesso!`))

	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao importar key...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
	
}

const keyAdd = async (args) => {
	console.log(chalk.gray(`Not implemented`));
}

const keyList = async (args) => {
	const table = new Table({
		head: ['Nome']
	});

	try {
		const r = await API.skipCache(args.cache === false).search(args._[2] || null).listKeys(args.g || null);

		if(!r.data || r.data.length == 0) {
			console.log(chalk.red(`Nenhuma chave encontrada`));
			process.exit(0);
		}
		
		for (const item of r.data) {
			table.push([
				// item.id,
				item.name
			]);
		}
		console.log(chalk.yellow.bold(`Chaves cadastradas:`));
		console.log(table.toString());

	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao listar...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(0);
	}
}

const keyDelete = async (args) => {
	const idOrAlias = args._[2] || null;
	if(!idOrAlias) {
		console.log(chalk.red(`ID ou Alias da chave é obrigatório, tente 'sshc key delete/del/rm [id OR name]'`));
		process.exit(1);
	}

	try {
		await API.deleteKey(idOrAlias);
		console.log(chalk.green.bold(`Chave excluida com sucesso!`));
	} catch(err){
		console.error(chalk.red('Ocorreu um erro ao deletar chave...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(0);
	}
}

const keyView = async (args) => {
	const idOrAlias = args._[2] || null;

	if(!idOrAlias) {
		console.log(chalk.red(`ID ou Name da chave é obrigatório, tente 'sshc key view [id OR name]'`));
		process.exit(1);
	}

	try {
		const r = await API.getKey(idOrAlias);

		if(!r.data || Object.keys(r.data).length == 0) {
			console.log(chalk.red(`Nenhuma chave encontrada`));
			process.exit(0);
		}

		console.log(chalk.cyan.bold(`Chave Privada:`));
		console.log(r.data.pvkey);
		console.log();

		if(r.data.keypass) {
			console.log(chalk.cyan.bold(`Senha da chave privada:`));
			console.log(r.data.keypass);
			console.log();
		}

		if(r.data.pubkey) {
			console.log(chalk.cyan.bold(`Chave Publica:`));
			console.log(r.data.pubkey);
			console.log();
		}

	} catch(err) {
		console.error(chalk.red('Ocorreu um erro...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(0);
	}
}

const serverAdd = async (args) => {

	const params = {
		name: args.n || args.name || null,
		alias: args.a || args.alias || null,
		host: args.h || args.host || null,
		port: args.p || args.port || 22,
		username: args.u || args.username || null,
		password: null,
		description: args.d || args.description || null,
		permId: args.k || null,
		groupId: args.g || null
	}

	// if(!params.name || typeof params.name != "string") {
	// 	console.log(chalk.red.bold(`Nome é obrigatório [-n|--name]`));
	// 	process.exit(1);
	// }

	if(!params.alias || typeof params.alias != "string") {
		console.log(chalk.red.bold(`Alias é obrigatória [-a|--alias]`));
		process.exit(1);
	}

	if(/[^a-z0-9\.\-_]/gi.test(params.alias)) {
		console.log(chalk.red.bold(`Alias inválido, é permitido apenas os caracteres [a-Z 0-9 - _ .]`));
		process.exit(1);
	}

	// if(!params.host || typeof params.host != "string") {
	// 	console.log(chalk.red.bold(`Host é obrigatório [-h|--host]`));
	// 	process.exit(1);
	// }

	if(params.host && /\:[0-9]{2,5}$/g.test(params.host)) {
		const [host,port] = params.host.split(':',2);
		params.host = host;
		params.port = Number(port);
	}

	// if(!params.port || typeof params.port != "number") {
	// 	console.log(chalk.red.bold(`Porta é obrigatória, tente ex.: [--port] OU ${params.host}:22 `));
	// 	process.exit(1);
	// }

	// if(!params.username || typeof params.username != "string") {
	// 	console.log(chalk.red.bold(`Username é obrigatório [-u|--username]`));
	// 	process.exit(1);
	// }

	if(args.pass || args.password) {
		const formData = await prompt({
			type: 'password',
			name: 'password',
			message: 'Digite a senha do servidor (ou deixe em branco):'
		});
		params.password = formData.password;
	}

	if(args.key) {
		try {
			const choices = { 'Não associar': null };
			const keys = await API.listKeys();
			for (const k of keys.data) {
				choices[k.name] = k.id;
			}

			const selectedKey = await new Select({
				name: 'chave',
				message: 'Selecione uma chave:',
				choices: Object.keys(choices)
			}).run();

			if(selectedKey && typeof choices[selectedKey] != "undefined") {
				params.permId = choices[selectedKey];
			}

		} catch(err) {
			console.error(chalk.red('Ocorreu um erro ao buscar chaves...'));
			if(err.response && err.response.data && err.response.data.message) 
				console.error(chalk.red(err.response.data.message));
			else
				console.error(chalk.red(err.message || err.stack || err));
			process.exit(1);
		}
	}

	if(args.group || args.grup) {
		try {
			const choices = { 'Não associar': null };
			const keys = await API.listGroups();
			for (const k of keys.data) {
				choices[k.name] = k.id;
			}
			const selectedOption = await new Select({
				name: 'grupo',
				message: 'Selecione um grupo:',
				choices: Object.keys(choices)
			}).run();

			if(selectedOption && typeof choices[selectedOption] != "undefined") {
				params.groupId = choices[selectedOption];
			}

		} catch(err) {
			console.error(chalk.red('Ocorreu um erro ao buscar grupos...'));
			if(err.response && err.response.data && err.response.data.message) 
				console.error(chalk.red(err.response.data.message));
			else
				console.error(chalk.red(err.message || err.stack || err));
			process.exit(1);
		}
	}

	try {
		const r = await API.newServer(params);
		console.log(chalk.green.bold(`Servidor salvo com sucesso!`));
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao salvar servidor...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}

}

const serverAddWithForm = async (args) => {
	console.log(chalk.gray(`Not implemented`));
	// const formData = await new Form({
	// 	name: 'server',
	// 	message: 'Adicionar um novo servidor:',
	// 	choices: [
	// 	  { name: 'name', message: 'Nome', initial: '', required: true },
	// 	  { name: 'alias', message: 'Alias/Atalho', initial: '', required: true, validate: s => !/[^a-z0-9\.\-_]/gi.test(s) },
	// 	  { name: 'host', message: 'Host', initial: '', required: true },
	// 	  { name: 'port', message: 'Porta', initial: '', required: true },
	// 	  { name: 'username', message: 'User', initial: '', required: true },
	// 	  { name: 'description', message: 'Descrição', initial: '', required: false }
	// 	]
	//   }).run();

	//   console.dir(formData);
}

const serverDelete = async (args) => {
	const idOrAlias = args._[2] || null;
	if(!idOrAlias) {
		console.log(chalk.red(`ID ou Alias do servidor é obrigatório, tente 'sshc server delete/del/rm [id OR alias]'`));
		process.exit(1);
	}

	try {
		await API.deleteServer(idOrAlias);
		console.log(chalk.green.bold(`Servidor excluido com sucesso!`));
	} catch(err){
		console.error(chalk.red('Ocorreu um erro ao deletar servidor...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		process.exit(1);
	}
}

const serverList = async (args) => {
	const table = new Table({
		head: ['Alias','Name', 'Username','Host/Port', 'Key', `Forward`] //(${chalk.green.bold('Local')}/${chalk.blue.bold('Remote')})
	});

	try {
		const r = await API.skipCache(args.cache === false).search(args._[2] || null).listServers(args.g || null);
		
		if(!r.data || r.data.length == 0) {
			console.error(chalk.red(`Nenhum servidor encontrado`));
			process.exit(0);
		}

		for (const item of r.data) {
			if(item.groupName) item.alias = `${item.alias} (${chalk.bold.cyan(item.groupName)})`;

			let forwardData = '';
			if(item.forwardOut && Object.keys(item.forwardOut).length > 0) {
				for (const i in item.forwardOut) {
					if (item.forwardOut.hasOwnProperty(i)) {
						const fwdText = item.forwardOut[i].replace('|',' → ').replace('0.0.0.0:','').replace('localhost:','').replace('127.0.0.1','');
						forwardData += `\n${chalk.green.bold('L')} ${fwdText}`;
					}
				}
			}
			if(item.forwardIn && Object.keys(item.forwardIn).length > 0) {
				for (const i in item.forwardIn) {
					if (item.forwardIn.hasOwnProperty(i)) {
						forwardData += `\n${chalk.blue.bold('R')} `+item.forwardIn[i].replace('|',' → ');
					}
				}
			}

			let hostField = item.port ? `${item.host}:${item.port}` : item.host;
			if(item.proxyServer && Object.keys(item.proxyServer).length > 0) hostField += `\n${chalk.magenta.bold('P →')} ${chalk.magenta(item.proxyServer.alias)}`;
			
			table.push([
				item.alias,
				item.name || '-',
				item.username,
				hostField,
				item.perm.name || '-',
				forwardData.trim() || '-'
			]);
		}

		console.log(chalk.yellow.bold(`Servidores cadastrados:`));
		console.log(table.toString());

	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao listar...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		if(args.debug) console.log(err.stack || err);
		process.exit(1);
	}
}

const serverAddHopping = async (args) => {
	const params = args._;

	if(!params[2] || !params[3]) {
		console.log(chalk.red(`Uso incorreto, tente 'sshc server hopping <server alias> <proxy server alias>'`));
		process.exit(1);
	}

	try {
		await API.addServerHopping(params[2], params[3]);
		console.log(chalk.green.bold(`Servidor associado com sucesso!`))
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

const serverDeleteHopping = async (args) => {
	const params = args._;

	if(!params[2]) {
		console.log(chalk.red(`Uso incorreto, tente 'sshc server hopping <server alias> --delete'`));
		process.exit(1);
	}

	try {
		await API.deleteServerHopping(params[2]);
		console.log(chalk.green.bold(`Servidor desassociado com sucesso!`))
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

const groupAddHopping = async (args) => {
	const params = args._;

	if(!params[2] || !params[3]) {
		console.log(chalk.red(`Uso incorreto, tente 'sshc group hopping <server alias> <proxy server alias>'`));
		process.exit(1);
	}

	try {
		await API.addGroupHopping(params[2], params[3]);
		console.log(chalk.green.bold(`Servidor associado com sucesso!`))
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

const groupDeleteHopping = async (args) => {
	const params = args._;

	if(!params[2]) {
		console.log(chalk.red(`Uso incorreto, tente 'sshc group hopping <server alias> --delete'`));
		process.exit(1);
	}

	try {
		await API.deleteGroupHopping(params[2]);
		console.log(chalk.green.bold(`Servidor desassociado com sucesso!`))
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

const serverForwardInOut = async (args) => {
	// sshc server forward localhost:9922 127.0.0.1:22
	const params = {
		server: args._[2],
		local: args._[3] || null,
		remote: args._[4] || null
	}

	if(!params.server || !params.local || !params.remote) {
		console.log(chalk.red(`Uso incorreto, tente 'sshc server forward <server alias> 9922 10.2.34.32:22'`));
		//* \nOU 'sshc server forward <server alias> 10.2.34.32:22' para encaminhamento remoto.
		process.exit(1);
	}

	if(!hostPortRegex.test(params.local) || !hostPortRegex.test(params.remote)) {
		console.log(chalk.red(`Uso incorreto, tente 'sshc server forward <server alias> 9922 10.2.34.32:22' [2]`));
		process.exit(1);
	}

	try {
		const r = await API.addForwardOutIn(params);
		console.log(chalk.green.bold(`Criado com sucesso!`));
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao salvar...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

const serverForwardInOutDelete = async (args) => {
	// sshc server forward localhost:9922 127.0.0.1:22
	const params = {
		server: args._[2],
		hash: null
	}

	try {
		const choices = { 'Cancelar': null };
		const forwards = await API.listForwardsOf(params.server).then(r => r.data);

		for (const i in forwards.forwardOut) {
			if (forwards.forwardOut.hasOwnProperty(i)) {
				const fwd = forwards.forwardOut[i];
				choices[`${chalk.green.bold('L')} ${fwd.replace('|',' → ')}`] = i;
			}
		}

		for (const i in forwards.forwardIn) {
			if (forwards.forwardIn.hasOwnProperty(i)) {
				const fwd = forwards.forwardIn[i];
				choices[`${chalk.blue.bold('R')} ${fwd}`] = i;
			}
		}

		const selectedOption = await new Select({
			name: 'forward',
			message: 'Selecione um item para deletar:',
			choices: Object.keys(choices)
		}).run();

		if(selectedOption && typeof choices[selectedOption] != "undefined") {
			params.hash = choices[selectedOption];
		}

	} catch(err) {
		if(err != '') {
			console.error(chalk.red('Ocorreu um erro.'));
			if(err.response && err.response.data && err.response.data.message) 
				console.error(chalk.red(err.response.data.message));
			else
				console.error(chalk.red(err.message || err.stack || err));
		}
		process.exit(1);
	}

	try {
		await API.deleteForwardOutIn(params.server, params.hash);
		console.log(chalk.green.bold(`Deletado com sucesso!`));
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao salvar...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

const groupAdd = async (args) => {
	const params = {
		name: args.n || args.name || null,
		port: args.p || args.port || null,
		username: args.u || args.username || null,
		password: null,
		description: args.d || args.description || null,
		permId: args.k || null
	}

	if(!params.name || typeof params.name != "string") {
		console.log(chalk.red.bold(`Nome é obrigatório [-n|--name]`));
		process.exit(1);
	}

	if(/[^a-z0-9\.\-_]/gi.test(params.name)) {
		console.log(chalk.red.bold(`Nome inválido, é permitido apenas os caracteres [a-Z 0-9 - _ .]`));
		process.exit(1);
	}

	if(args.pass || args.password) {
		const formData = await prompt({
			type: 'password',
			name: 'password',
			message: 'Digite a senha do servidor (ou deixe em branco):'
		});
		params.password = formData.password;
	}

	if(args.key) {
		try {
			const choices = { 'Não associar': null };
			const keys = await API.listKeys();
			for (const k of keys.data) {
				choices[k.name] = k.id;
			}

			const selectedKey = await new Select({
				name: 'chave',
				message: 'Selecione uma chave:',
				choices: Object.keys(choices)
			}).run();

			if(selectedKey && typeof choices[selectedKey] != "undefined") {
				params.permId = choices[selectedKey];
			}

		} catch(err) {
			console.error(chalk.red('Ocorreu um erro ao buscar chaves...'));
			if(err.response && err.response.data && err.response.data.message) 
				console.error(chalk.red(err.response.data.message));
			else
				console.error(chalk.red(err.message || err.stack || err));
			process.exit(1);
		}
	}

	try {
		const r = await API.newGroup(params);
		console.log(chalk.green.bold(`Grupo salvo com sucesso!`));
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao salvar...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

const groupList = async (args) => {
	const table = new Table({
		head: ['Nome','Porta','Usuário','Key', '-']
	});

	try {
		const r = await API.skipCache(args.cache === false).search(args._[2] || null).listGroups();
		
		if(!r.data || r.data.length == 0) {
			console.log(chalk.red(`Nenhum grupo encontrado`));
			process.exit(0);
		}
		
		for (const item of r.data) {
			let addinfo = '';
			if(item.proxyServer && item.proxyServer.alias) addinfo += `${chalk.magenta.bold('P →')} ${chalk.magenta(item.proxyServer.alias)}`
			table.push([
				item.name,
				item.port || '-',
				item.username || '-',
				item.perm.name || '-',
				addinfo || '-'
			]);
		}
		console.log(chalk.yellow.bold(`Grupos cadastrados:`));
		console.log(table.toString());

	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao listar...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(0);
	}
}

const groupDelete = async (args) => {
	const idOrAlias = args._[2] || null;
	if(!idOrAlias) {
		console.log(chalk.red(`Nome do grupo é obrigatório, tente 'sshc group delete/del/rm <group name>'`));
		process.exit(1);
	}

	try {
		await API.deleteGroup(idOrAlias);
		console.log(chalk.green.bold(`Grupo excluido com sucesso!`));
	} catch(err){
		console.error(chalk.red('Ocorreu um erro ao deletar...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(0);
	}
}

const groupUnlink = async (args) => {
	const params = args._;

	if(!params[2]) {
		console.log(chalk.red.bold(`Uso incorreto, tente 'sshc group unlink <server alias>'`));
		process.exit(1);
	}

	try {
		await API.unlinkGroup(params[2]);
		console.log(chalk.green.bold(`Grupo desassociado com sucesso!`))
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao desassociar...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

const groupLink = async (args) => {
	const params = args._;

	if(!params[2] || !params[3]) {
		console.log(chalk.red.bold(`Uso incorreto, tente 'sshc group <group name> <server alias>'`));
		process.exit(1);
	}
	
	try {
		await API.linkGroup(params[2], params[3]);
		console.log(chalk.green.bold(`Group associado com sucesso!`))
	} catch(err) {
		console.error(chalk.red('Ocorreu um erro ao associar...'));
		if(err.response && err.response.data && err.response.data.message) 
			console.error(chalk.red(err.response.data.message));
		else
			console.log(chalk.red(err.message || 'unknown'));
		
		process.exit(1);
	}
}

module.exports = {
	initCLI,
	connect,
	clearCache,
	purge,
	debug,
	
	validate,
	refresh,
	register,
	login,
	register,
	relogin,
	loginMFA,
	logout,
	removePhoneNumber,
	registerPhoneNumber,
	statusMFA,
	enableMFA,
	disableMFA,

	keyUnlink,
	keyLink,
	keyImport,
	keyAdd,
	keyList,
	keyDelete,
	keyView,

	serverAdd,
	serverAddWithForm,
	serverDelete,
	serverList,
	serverForwardInOut,
	serverForwardInOutDelete,
	serverAddHopping,
	serverDeleteHopping,

	groupAdd,
	groupList,
	groupDelete,
	groupUnlink,
	groupLink,
	groupAddHopping,
	groupDeleteHopping
}