/*
 * Description: Interactive SSH Terminal running on nodejs
 * Author: Aaron Gong
 * Date: 2014-11-17
 * http://www.pulse.sg/posts/nodejs_ssh_terminal
 * Licence: MIT
 */
const wsize = require('window-size');
const chalk = require('chalk');
const net = require('net');
const getPort = require('get-port');
const Connection = require('ssh2');
const conn = new Connection();

let windowSize = wsize.get();

const forwardIn = (host, port) => {
	return new Promise((resolve, reject) => {
		conn.forwardIn(host, port, function(err) {
			if (err) reject(err);
			else resolve(true)
		});
	})
}

const forwardOut = (srcAddr, srcPort, dstAddr, dstPort) => {
	return new Promise((resolve, reject) => {
		net.createServer((socket) => {
			conn.forwardOut(srcAddr, srcPort, dstAddr, dstPort, (err, streamFO) => {
				if(err) reject(err);
				else {
					streamFO.pause();
					socket.pipe(streamFO)
					streamFO.pipe(socket)
					streamFO.resume();
					streamFO.on('close',() => console.log(chalk.magenta(`ForwardOut closed ${srcAddr}:${srcPort} → ${dstAddr}:${dstPort}`)))
				}
			})
		}).listen(+srcPort, srcAddr,() => {
			resolve();
		}).on('error', err => {
			reject(err);
		})
	})
}

const createProxyConnection = (ops, target) => {
	return new Promise((resolve, reject) => {
		const connProxy =  new Connection();

		connProxy.on('error', (err) => {
			if(err.level == 'client-authentication') // client-timeout
				console.error(chalk.red(`Ocorreu um erro na autenticação do proxy server. ${ops.username}@${ops.host}:${ops.port}`));
			else
				console.error(chalk.red(err.message),chalk.gray(err.level));
			
			reject(err);
		});

		connProxy.on('ready', async function () {
			const port = await getPort();
			connProxy.forwardOut('localhost', port, target.host, target.port, (err, stream) => {
				if(err) reject(err)
				else resolve(stream);
			});
		})

		connProxy.connect(ops);
	})
}

module.exports = (ops) => {
	return new Promise(async (resolve, reject) => {
		let gs = null;

		if(ops.proxyServer) {
			console.log(chalk.gray(`Conectando via ${ops.proxyServer.alias}...`));
			ops.sock = await createProxyConnection(ops.proxyServer, ops);
		}

		conn.on('error', (err) => {
			if(err.level == 'client-authentication') // client-timeout
				console.error(chalk.red(`Ocorreu um erro na autenticação. ${ops.username}@${ops.host}:${ops.port}`));
			else
				console.error(chalk.red(err.message),chalk.gray(err.level));
			
			process.exit('SIGKILL');
		});

		// conn.on('connect', function() {});

		conn.on('ready', async function () {
			// console.log(chalk.yellow('Connection :: ready!'));

			// ForwardOut
			if(ops.forwardOut && ops.forwardOut.length > 0)
				for (const i in ops.forwardOut) {
					if (ops.forwardOut.hasOwnProperty(i)) {
						const fwd = ops.forwardOut[i];
						const inPort = await getPort(fwd.in);
						await forwardOut(fwd.in.host, inPort, fwd.out.host, fwd.out.port)
						.then(() => console.log(chalk.cyan(` - ForwardOut ${fwd.in.host}:${inPort == fwd.in.port ? inPort : chalk.magenta(inPort)} → ${fwd.out.host}:${fwd.out.port}`)))
						.catch(err => console.error(chalk.red(err.message) || err));
					}
				}
			
			// ForwardIn
			// if(ops.forwardIn && ops.forwardIn.length > 0)
			// 	for (const i in ops.forwardIn) {
			// 		if (ops.forwardIn.hasOwnProperty(i)) {
			// 			const fwd = ops.forwardIn[i];
			// 			await forwardIn(fwd.host, fwd.port)
			// 			.then(() => console.log(chalk.magenta(` - ForwardIn ${fwd.host}:${fwd.port}`)))
			// 			.catch(err => console.error(chalk.red(err.message) || err));
			// 		}
			// 	}

			conn.shell(async function (err, stream) {
				if (err) reject(err);
				else {
					// Set initial window size
					stream.setWindow(process.stdout.rows, process.stdout.columns, windowSize.height, windowSize.width);

					stream.on('close', function () {
						// console.log(chalk.cyan('Connection :: closed'));
						conn.end();
						process.exit(0);
					}).on('data', function (data) {
						if (!gs) gs = stream;
						// if (gs._writableState.sync == false) 
						process.stdout.write('' + data);
					}).stderr.on('data', function (data) {
						// console.log('STDERR: ' + data);
						process.stderr.write('' + data);
					});

					// Set window size on resizing
					process.stdout.on('resize', () => {
						windowSize = wsize.get();
						stream.setWindow(process.stdout.rows, process.stdout.columns, windowSize.height, windowSize.width);
					});

					resolve(true);
				}
			});

		}).connect(ops);

		// The code below is from the link below. I just sent key press to stream used by SSH instead of stdout
		// http://stackoverflow.com/questions/5006821/nodejs-how-to-read-keystrokes-from-stdin
		// 
		const stdin = process.stdin;
		stdin.setRawMode(true); // without this, we would only get streams once enter is pressed
		stdin.resume(); // resume stdin in the parent process (node app won't quit all by itself unless an error or process.exit() happens)
		stdin.setEncoding('utf8'); // i don't want binary, do you?
		stdin.on('data', function (key) { // on any data into stdin
			if (key === '\u001a' || (!gs && key === '\u0003') ) { // ctrl-c ( end of text ) u0004
				process.exit('SIGKILL');
			}
			if (gs) gs.write('' + key); // write the key to stdout all normal like
		});
	})
}