const config = require('../bin/config');
const chalk = require('chalk');
const axios = require('axios');
const crypto = require('crypto');
const cache = require('../lib/Cache');

const sha1 = s => crypto.createHash('sha1').update(s).digest('hex');
const APIURL = !process.env.RANIENV ? 'https://api-cli.sshchain.com.br' : 'http://localhost:3038';
const CACHE_TTL = 300;

class API {

    static create(...args) {
        return new API(...args);
    }

    constructor() {
        this.wordSearch = null;
        this.cacheBypass = false;

        this._req = axios.create({
            baseURL: APIURL,
            method: 'POST',
            responseType: 'json',
            headers: {
                'Content-Type': 'application/json;charset=utf-8',
                'Cache-Control': 'no-cache'
            }
       });
    }

    getForConnect(serverAlias) {
        const ops = {
            method: 'GET',
            url: `/chain/connect`,
            params: {
                alias: serverAlias
            },
            data: {},
            cacheTTL: 60 * 60 * 24 * 1
        };
        return this.req(ops);
    }

    getKey(idOrName) {
        const ops = {
            method: 'GET',
            url: `/chain/pem/full`,
            params: {
                key: idOrName
            }
        };
        return this.req(ops);
    }

    deleteKey(keyAliasOrId) {
        const ops = {
            method: 'DELETE',
            url: `/chain/pem`,
            data: {
                key: keyAliasOrId
            }
        };
        return this.req(ops);
    }

    newKey(nameFile, privateFile, privateKeyPassword = null, publicFile = null) {
        const ops = {
            method: 'POST',
            url: `/chain/pem`,
            data: {
                pv: privateFile,
                pb: publicFile,
                name: nameFile,
                pass: privateKeyPassword
            }
        };
        return this.req(ops);
    }

    listKeys(group = null) {
        const ops = {
            method: 'GET',
            url: `/chain/pem`,
            params: {
                group: group,
                search: this.wordSearch || null
            },
            cacheTTL: CACHE_TTL
        };
        return this.req(ops);
    }

    linkKey(keyAlias, target = {}) {
        const ops = {
            method: 'POST',
            url: `/chain/pem/link`,
            data: Object.assign(target, { key: keyAlias })
        };
        return this.req(ops);
    }

    unlinkKey(serverAlias = null, groupName = null) {
        const ops = {
            method: 'POST',
            url: `/chain/pem/unlink`,
            data: {
                server: serverAlias,
                group: groupName
            }
        };
        return this.req(ops);
    }

    listServers(group = null) {
        const ops = {
            method: 'GET',
            url: `/chain/server`,
            params: {
                group: group,
                search: this.wordSearch || null
            },
            cacheTTL: CACHE_TTL
        };
        return this.req(ops);
    }

    newServer(body) {
        const ops = {
            method: 'POST',
            url: `/chain/server`,
            data: body
        };
        return this.req(ops);
    }

    deleteServer(idOrAlias) {
        const ops = {
            method: 'DELETE',
            url: `/chain/server`,
            data: {
                key: idOrAlias
            }
        };
        return this.req(ops);
    }

    addForwardOutIn(body) {
        const ops = {
            method: 'POST',
            url: `/chain/server/forward`,
            data: body
        };
        return this.req(ops);
    }

    listForwardsOf(idOrAlias) {
        const ops = {
            method: 'GET',
            url: `/chain/server/forward`,
            data: {
                server: idOrAlias
            }
        };
        return this.req(ops);
    }

    deleteForwardOutIn(server, fwdHash) {
        const ops = {
            method: 'DELETE',
            url: `/chain/server/forward`,
            data: {
                server: server,
                hash: fwdHash
            }
        };
        return this.req(ops);
    }

    addServerHopping(server, serverProxy){
        const ops = {
            method: 'POST',
            url: `/chain/server/hopping`,
            data: {
                server: server,
                proxy: serverProxy
            }
        };
        return this.req(ops);
    }

    deleteServerHopping(server){
        const ops = {
            method: 'DELETE',
            url: `/chain/server/hopping`,
            data: {
                server: server
            }
        };
        return this.req(ops);
    }

    addGroupHopping(group, serverProxy){
        const ops = {
            method: 'POST',
            url: `/chain/group/hopping`,
            data: {
                group: group,
                proxy: serverProxy
            }
        };
        return this.req(ops);
    }

    deleteGroupHopping(group){
        const ops = {
            method: 'DELETE',
            url: `/chain/group/hopping`,
            data: {
                group: group
            }
        };
        return this.req(ops);
    }

    listGroups() {
        const ops = {
            method: 'GET',
            url: `/chain/group`,
            params: {
                search: this.wordSearch || null
            },
            cacheTTL: CACHE_TTL
        };
        return this.req(ops);
    }

    newGroup(options) {
        const ops = {
            method: 'POST',
            url: `/chain/group`,
            data: options
        };
        return this.req(ops);
    }

    deleteGroup(idOrAlias) {
        const ops = {
            method: 'DELETE',
            url: `/chain/group`,
            data: {
                key: idOrAlias
            }
        };
        return this.req(ops);
    }

    linkGroup(groupName, serverAlias) {
        const ops = {
            method: 'POST',
            url: `/chain/group/link`,
            data: {
                group: groupName,
                server: serverAlias
            }
        };
        return this.req(ops);
    }

    unlinkGroup(serverAlias = null) {
        const ops = {
            method: 'POST',
            url: `/chain/group/unlink`,
            data: {
                server: serverAlias
            }
        };
        return this.req(ops);
    }

    skipCache(v = true) {
        this.cacheBypass = v;
        return this;
    }

    search(word) {
        if(word) this.wordSearch = word;
        return this;
    }

    checkConnection() {
        return new Promise((resolve, reject) => {
            require('dns').lookup('api-cli.sshchain.com.br',err => {
                if (err && err.code == "ENOTFOUND")
                    resolve(false);
                else
                    resolve(true);
            })
        })
    }

    async req(ops, useToken = true) {

        const sha1SUM = sha1(JSON.stringify(ops));
        const ttl = ops.cacheTTL || 0;

        if(useToken && config.has('tokenData.accessToken')) {
            if(!ops.headers) ops.headers = {};
            ops.headers['accessToken'] = config.get('tokenData.accessToken');
        }

        if(!this.cacheBypass && ttl > 0 && cache.hasValid(sha1SUM, ttl)) {
            console.log(chalk.gray(`Carregando informações do cache, use --no-cache para pular/atualizar o cache.`));
            return cache.get(sha1SUM);
        }
        
        return this._req(ops).then(r => r.data || r).then(r => {
            if(ttl > 0) cache.set(sha1SUM,r,ttl);
            return r;
        });
    }
    
}

module.exports = API.create();