const config = require('../bin/config');
const fs = require('fs');
const path = require('path');
const crypto = require("crypto");

const sha1 = s => crypto.createHash('sha1').update(s).digest('hex');
const getTime = () => new Date().getTime();
const random = () => ~~(Math.random() * 1000) + 1;

class Cache {

    static create(...args) {
        return new Cache(...args);
    }

    constructor(options = {}) {
        this.cachePath = options.cachePath ? options.cachePath : (config.get('appUserDir') || './diskcache');
        if(!fs.existsSync(this.cachePath)) fs.mkdirSync(this.cachePath);
        if(!path.isAbsolute(this.cachePath)) this.cachePath = fs.realpathSync(this.cachePath);
        if(random() % 8 == 0) this.clearExpired();
    }

    has(key) {
        const fullPath = path.join(this.cachePath, `${key}.json`);
        return fs.existsSync(fullPath);
    }

    hasValid(key, ttl = 0) {
        const fullPath = path.join(this.cachePath, `${key}.json`);
        if(!fs.existsSync(fullPath)) return false;
        if(ttl == 0) return true;
        const stats = fs.statSync(fullPath);
        const validateTime = getTime() <= (stats.birthtimeMs + (ttl * 1000))
        if(!validateTime) fs.unlinkSync(fullPath);
        return validateTime;
    }

    get(key, ttl = 0) {
        const fullPath = path.join(this.cachePath, `${key}.json`);
        if(!fs.existsSync(fullPath)) return null;
        let fileData = fs.readFileSync(fullPath, { encoding: 'utf-8' });
            fileData = JSON.parse(fileData);
        if(getTime() >= fileData.expiresOn) fs.unlinkSync(fullPath);
        return fileData.data || null;
    }

    set(key, value, ttl = 0) {
        const fullPath = path.join(this.cachePath, `${key}.json`);
        const dataSave = {
            expiresOn: getTime() + (ttl * 1000),
            isObject: (typeof value == "object"),
            data: value
        };
        return fs.writeFileSync(fullPath, JSON.stringify(dataSave), { encoding: 'utf-8' });
    }
    
    remove(key) {
        const fullPath = path.join(this.cachePath, `${key}.json`);
        if(!fs.existsSync(fullPath)) return true;
        return fs.unlinkSync(fullPath);
    }

    clearAll() {
        const filesList = fs.readdirSync(this.cachePath).filter(f => f.endsWith('.json'));
        if(filesList.length == 0) return true;
        for (const file of filesList) {
            fs.unlinkSync(path.join(this.cachePath,file));
        }
        return true;
    }

    async clearExpired() {
        try {
            const filesList = fs.readdirSync(this.cachePath).filter(f => f.endsWith('.json'));
            if(filesList.length == 0) return true;
            for (const file of filesList) {
                const filePath = path.join(this.cachePath,file);
                const fileData = await fs.readFileSync(filePath).toJSON() || {};
                if(getTime() >= fileData.expiresOn) fs.unlinkSync(filePath);
            }
            return true;
        } catch(err) {
            return false;
        }
    }

    performPersist() {
        return true;
    }
}

module.exports = Cache.create();